import os
import shutil

import lz4.frame
from farmi.serialization import deserialize_file

ONE_GB = 1_073_741_824


class FarmiFile(object):
    """handles saving to a farmi file"""

    def __init__(self, file_name):
        self.file_name = file_name
        self.file_handle = None
        self.counter = 0
        self.file_timestamp = None
        self.should_create_new_file = True
        self.ready_to_compress = []

    def check_file_size(self):
        if self.file_handle and os.path.getsize(self.file_handle.name) > ONE_GB * 5:
            self.should_create_new_file = True

            compress_files = self.ready_to_compress[:]
            self.ready_to_compress = []

            for filename in compress_files:
                new_path = filename + ".lz4"
                with open(filename, "rb") as f_in, lz4.frame.open(
                        new_path, "wb"
                ) as f_out:
                    shutil.copyfileobj(f_in, f_out)
                    os.remove(f_in.name)

    def write(self, data=None, timestamp=None):
        if self.should_create_new_file:
            self._create_new_file(timestamp)

        self.file_handle.write(data)
        # not sure if we should flush each time..
        # self.file_handle.flush()

    def _create_new_file(self, timestamp=None):
        if not self.file_timestamp:
            self.file_timestamp = timestamp

        self.counter += 1
        if self.file_handle:
            self.ready_to_compress.append(self.file_handle.name)
            self.file_handle.close()

        self.file_handle = open(
            self.file_name.format(f"{self.file_timestamp}_{self.counter}"), "wb"
        )
        self.should_create_new_file = False

    @staticmethod
    def read(filename):
        with open(filename, "rb") as f:
            for value in deserialize_file(f):
                yield value

    def close(self):
        self.file_handle.close()



import json
import re
from collections import defaultdict

import zmq

from farmi.farmi import Farmi
from farmi.serialization import deserialize


class Subscriber(Farmi):
    def __init__(self, directory_service_address="tcp://127.0.0.1:5555"):
        super().__init__(directory_service_address)
        self.topics = defaultdict(dict)
        self.poller = zmq.Poller()

        self.directory_service.send_json({"action": "GET_PUB_ADDRESS"})
        sub_address = self.directory_service.recv_string()

        self.directory_service_sub = self.context.socket(zmq.SUB)
        self.directory_service_sub.connect(sub_address)
        self.directory_service_sub.subscribe("")

        self.poller.register(self.directory_service_sub, zmq.POLLIN)

    def get_matching_topics(self, topic):
        for topic_matchers in self.topics.keys():
            if topic_matchers.match(topic):
                yield topic_matchers

    def subscribe_to(self, topic, fn):
        try:
            topic.match("")
        except:
            topic = re.compile("^%s$" % re.escape(topic))
        self.topics[topic]["fn"] = fn

        self.directory_service.send_json({"action": "TOPICS"})
        topics = self.directory_service.recv_json()
        for received_topic, info in topics.items():
            for matching_topic in self.get_matching_topics(received_topic):
                self._add_topic(matching_topic, info["address"])

    def _add_topic(self, topic, address):
        self.topics[topic]["address"] = address
        socket = self.context.socket(zmq.SUB)
        socket.connect(address)
        socket.subscribe("")
        self.topics[topic]["socket"] = socket
        self.poller.register(self.topics[topic]["socket"], zmq.POLLIN)

    def _remove_topic(self, topic):
        if self.topics[topic].get("socket"):
            self.poller.unregister(self.topics[topic]["socket"])
            self.topics[topic]["socket"].close()
            self.topics[topic].pop("socket")
            self.topics[topic].pop("address")

    def close(self):
        for topic, info in self.topics.items():
            if info["socket"]:
                info["socket"].close()
        super().close()

    def listen(self):
        while not self.exit.is_set():
            poller = dict(self.poller.poll())
            for poll in poller:
                raw_msg = poll.recv_multipart()
                if poll == self.directory_service_sub:
                    msg = json.loads(raw_msg[1].decode("utf-8"))
                    action = msg.get("action")
                    if action == "REGISTERED":
                        topic, address = msg.get("topic"), msg.get("address")
                        for matching_topic in self.get_matching_topics(topic):
                            if self.topics[matching_topic].get("address") != address:
                                self._add_topic(matching_topic, address)
                    elif action == "DEREGISTERED":
                        for matching_topic in list(
                            self.get_matching_topics(msg.get("topic"))
                        )[:]:
                            self._remove_topic(matching_topic)
                    for matching_topic in self.get_matching_topics(action):
                        self.topics[matching_topic]["fn"](
                            msg.get("topic"), msg.get("time"), msg
                        )
                else:
                    for topic in self.topics.values():
                        if topic.get("socket") == poll:
                            if len(raw_msg) == 3:
                                try:
                                    topic_, time_, body = raw_msg
                                except ValueError as e:
                                    print(raw_msg)
                                    raise e
                                deserialized_body = deserialize(body)
                                if isinstance(
                                    deserialized_body, list
                                ) and deserialized_body[0] == topic_.decode("utf-8"):
                                    b = deserialized_body[2]
                                else:
                                    b = deserialized_body
                                topic["fn"](
                                    topic_.decode("utf-8"),
                                    float(time_.decode("utf-8")),
                                    b,
                                )



import time
from threading import Thread

import zmq

from farmi.farmi import Farmi
from farmi.farmi_file import FarmiFile
from farmi.serialization import serialize
from farmi.utils import get_ip


class Publisher(Farmi):
    def __init__(
        self,
        topic,
        local_save=None,
        directory_service_address="tcp://127.0.0.1:5555",
        heartbeat_frequency=5,
        compress_data=None,
    ):
        super().__init__(directory_service_address)
        self.topic = topic
        self.heartbeat_frequency = heartbeat_frequency
        self.pub_address = None
        self.pub_socket = self.context.socket(zmq.PUB)
        self.local_save = local_save
        self.compress_data = compress_data
        self._create_publisher()
        Thread(target=self._heartbeat).start()

        self.file_handle = self._create_farmi_file()

    def _create_farmi_file(self):
        if self.local_save:
            dir_name = self.local_save if isinstance(self.local_save, str) else "."
            return FarmiFile(f"{dir_name}/{self.topic}-{{}}.farmi")
        else:
            return None

    def _heartbeat(self):
        while not self.exit.is_set():
            self.directory_service.send_json(
                {"action": "HEARTBEAT", "topic": self.topic, "time": time.time()}
            )
            response = self.directory_service.recv_string()
            if response == "NOT_REGISTERED":
                self._register()
            if self.compress_data and self.file_handle:
                self.file_handle.check_file_size()
            self.exit.wait(self.heartbeat_frequency)

    def _register(self):
        self.directory_service.send_json(
            {"action": "REGISTER", "topic": self.topic, "address": self.pub_address}
        )
        self.directory_service.recv_string()

    def _create_publisher(self):
        zmq_port = self.pub_socket.bind_to_random_port("tcp://*", max_tries=150)
        self.pub_address = "tcp://{}:{}".format(get_ip(), zmq_port)
        self._register()

    def end_recording(self):
        if self.file_handle:
            self.file_handle.close()
            self.file_handle = self._create_farmi_file()

    def send(self, data, timestamp=None):
        if not timestamp:
            timestamp = time.time()
        data_time = timestamp + self.time_offset

        serialized_data = serialize([self.topic, data_time, data])

        if self.file_handle:
            self.file_handle.write(data=serialized_data, timestamp=data_time)

        self.pub_socket.send_multipart(
            (
                self.topic.encode("utf-8"),
                str(data_time).encode("utf-8"),
                serialized_data,
            )
        )

    def close(self):
        self.directory_service.send_json({"action": "DEREGISTER", "topic": self.topic})
        self.directory_service.recv()
        super().close()
        self.pub_socket.send(b"CLOSE")
        self.pub_socket.close()
        if self.file_handle:
            self.file_handle.close()



import json
import logging
import time

import zmq

from farmi.utils import get_ip

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class DirectoryService(object):
    def __init__(self, port=5555, purge_timeout=20):
        self._topics = {}
        self.purge_timeout = purge_timeout
        context = zmq.Context()
        self.rep_socket = context.socket(zmq.REP)
        self.rep_socket.bind("tcp://*:{}".format(port))

        self.publisher = context.socket(zmq.PUB)
        pub_port = self.publisher.bind_to_random_port("tcp://*", max_tries=150)
        self.pub_address = "tcp://{}:{}".format(get_ip(), pub_port)

    @property
    def topics(self):
        self._purge_topics()
        return self._topics

    def _purge_topics(self):
        to_be_deleted = []
        for topic, info in self._topics.items():
            if time.time() - info["latest_heartbeat"] > self.purge_timeout:
                to_be_deleted.append(topic)

        for topic in to_be_deleted:
            self._deregister(topic)

    def _deregister(self, topic):
        if self._topics.get(topic):
            self._topics.pop(topic)
            self._broadcast(json.dumps({"action": "DEREGISTERED", "topic": topic}))

    def listen(self):
        while True:
            raw_msg = self.rep_socket.recv_string()
            try:
                msg = json.loads(raw_msg)
            except ValueError:
                logger.info("MALFORMED_JSON" + raw_msg)
                self.rep_socket.send_string("MALFORMED_JSON")
                continue

            action = msg.get("action")

            if action == "GET_PUB_ADDRESS":
                response = self.handle_get_pub_address(msg)
            elif action == "REGISTER":
                response = self.handle_register(msg)
            elif action == "DEREGISTER":
                response = self.handle_deregister(msg)
            elif action == "SYNC":
                response = self.handle_sync(msg)
            elif action == "ANNOUNCE":
                response = self.handle_announce(msg)
            elif action == "HEARTBEAT":
                response = self.handle_heartbeat(msg)
            elif action == "TOPICS":
                response = self.handle_topics(msg)
            else:
                response = "COMMAND_NOT_FOUND"

            logger.info("received msg: {}, responded: {}".format(msg, response))
            self.rep_socket.send_string(response)

    def _broadcast(self, data, topic=""):
        logger.info("broadcasting: {}".format(data))
        self.publisher.send_multipart([topic.encode("utf-8"), data.encode("utf-8")])

    def _check_required_args(self, msg, required):
        return all([msg.get(x) for x in required])

    def handle_topics(self, msg):
        return json.dumps(self.topics)

    def handle_get_pub_address(self, msg):
        return self.pub_address

    def handle_deregister(self, msg):
        if not self._check_required_args(msg, ["topic"]):
            return "MISSING_INFORMATION"
        self._deregister(msg.get("topic"))
        return "OK"

    def handle_register(self, msg):
        if not self._check_required_args(msg, ["topic", "address"]):
            return "MISSING_INFORMATION"

        topic, address = msg.get("topic"), msg.get("address")

        if self.topics.get(topic):
            self._deregister(topic)
            # return 'TOPIC_ALREADY_REGISTERED'

        self.topics[topic] = {"address": address, "latest_heartbeat": time.time()}
        self._broadcast(
            json.dumps({"action": "REGISTERED", "topic": topic, "address": address})
        )
        return "OK"

    def handle_sync(self, msg):
        return str(time.time())

    def handle_announce(self, msg):
        if not self._check_required_args(msg, ["msg"]):
            return "MISSING_INFORMATION"

        self._broadcast(msg.get("msg"))
        return "OK"

    def handle_heartbeat(self, msg):
        if not self._check_required_args(msg, ["topic", "time"]):
            return "MISSING_INFORMATION"

        logger.info("got HEARTBEAT: {}".format(msg))

        topic, time_ = msg.get("topic"), msg.get("time")

        if self._topics.get(topic):
            self._topics[topic]["latest_heartbeat"] = time.time()

            self._broadcast(
                json.dumps(
                    {
                        "action": "HEARTBEAT",
                        "topic": topic,
                        "time": time.time(),
                        "sender_time": time_,
                    }
                )
            )
            return "OK"
        else:
            return "NOT_REGISTERED"



import unittest
from io import BytesIO

import numpy as np

from farmi.serialization import deserialize, deserialize_file, serialize


class TestSerialization(unittest.TestCase):
    def test_serializer(self):
        data = ["some strange stuff!", {"okay": ["coolio", "woho"]}, 4.44]

        serialized_data = serialize(data)

        deserialized_data = deserialize(serialized_data)

        self.assertEqual(data, deserialized_data)

    def test_serialization_of_numpy(self):
        data = np.random.random((10, 15, 20))

        serialized_data = serialize(data)
        deserialized_data = deserialize(serialized_data)

        self.assertTrue((data == deserialized_data).all())

    def test_serialization_of_numpy_in_file_serializer(self):
        data = np.random.random((10, 15, 20))

        outfile = BytesIO()
        outfile.write(serialize(("t", 1.2, data)))
        outfile.seek(0)

        deserialized_data = list(deserialize_file(outfile))

        self.assertTrue((data == deserialized_data[0][2]).all())

    def test_file_serializer(self):
        data = ["topic1", 4.231, ["some strange stuff!", {"okay": "coolio"}, 4.44]]
        data2 = ["topic2", 1.231, 2.312]
        data3 = ["topic3", 9.23, {2.0: ["list of things"]}]

        outfile = BytesIO()

        outfile.write(serialize(data))
        outfile.write(serialize(data2))
        outfile.write(serialize(data3))

        outfile.seek(0)

        deserialized_data = list(deserialize_file(outfile))

        self.assertEqual(data, deserialized_data[0])
        self.assertEqual(data2, deserialized_data[1])
        self.assertEqual(data3, deserialized_data[2])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name="farmi",
    install_requires=["pyzmq", "msgpack", "msgpack-numpy"],
    python_requires=">=3",
    version="3.1.2",
    description="Multisensory processing framework ",
    author="Patrik Jonell",
    author_email="pjjonell@kth.se",
    packages=find_packages(),
    scripts=["bin/farmi-server"],
)



import time
from threading import Event

import zmq


class Farmi(object):
    """docstring for Farmi"""

    def __init__(self, directory_service_address="tcp://127.0.0.1:5555"):
        self.exit = Event()
        self.context = zmq.Context()
        self.directory_service = self.context.socket(zmq.REQ)
        self.directory_service.connect(directory_service_address)
        self._sync_time()

    def _sync_time(self):
        self.time_offset = 0
        number_of_syncs = 3

        for i in range(number_of_syncs):
            self.directory_service.send_json({"action": "SYNC"})
            time_offset = float(self.directory_service.recv_string()) - time.time()
            self.time_offset += time_offset

        self.time_offset /= number_of_syncs

    def close(self):
        self.exit.set()
        time.sleep(0.3)
        self.directory_service.close()



import msgpack
import msgpack_numpy as m

m.patch()


def deserialize_file(f, legacy=False):
    return msgpack.Unpacker(f, raw=legacy, use_list=True, max_str_len=99999999999)


def deserialize(data):
    return msgpack.unpackb(data, raw=False)


def serialize(data):
    return msgpack.packb(data, use_bin_type=True)



import socket

from farmi.serialization import deserialize_file


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        ip = s.getsockname()[0]
    except:
        ip = "127.0.0.1"
    finally:
        s.close()
    return ip


def read_farmi_file(filename):
    with open(filename, "rb") as f:
        for value in deserialize_file(f):
            yield value
